/*******************************************************************************

   KEYBOARD_4x4_PS2


   Objetivo: Usando protoclo PS2 (teclado) enviar cdigo de teclas para o PC
             utilizando um simples teclado matricial de telefone convencional.
             Estas teclas so definidas no programa e podem ser alteradas.

   Autor: Cludio Larios

   Data:  28/12/2016

   Uso didtico
   
   Nota: Testado em PC com sistema operacional XP.
   No sei se operar com outros operacionais mais modernos.

   Este arquivo  parte integrante do blog Larios.tecnologia.ws

******************************************************************************/
#include <16F628A.h>
#use delay(clock=4000000)
#fuses NOWDT,INTRC_IO, NOPUT, NOPROTECT, BROWNOUT, MCLR, NOLVP, NOCPD



//byte
#byte  porta       = 0x05
#byte  trisa       = 0x85
#byte  portb       = 0x06
#byte  trisb       = 0x86
//bit
#bit   clk         = 0x06.0  //pino 6
#bit   dado        = 0x05.4  //pino 3
#bit   tris_clk    = 0x86.0
#bit   tris_dado   = 0x85.4
//#bit   sw          = 0x06.3  //pino 9  //teste

//variaveis globais
unsigned int8 buf,sig;
unsigned int8 tecla=0xff;
unsigned int8 old_tecla=0xff;
unsigned int1 f_tec=0;

//codes a ser enviado pelas teclas ao soltar o teclado
#define tecla_0_break  Kbd_send(0xf0); Kbd_send(0x45);//code para letra  '0'
#define tecla_1_break  Kbd_send(0xf0); Kbd_send(0x16);
#define tecla_2_break  Kbd_send(0xf0); Kbd_send(0x1e);
#define tecla_3_break  Kbd_send(0xf0); Kbd_send(0x26);
#define tecla_4_break  Kbd_send(0xf0); Kbd_send(0x25);
#define tecla_5_break  Kbd_send(0xf0); Kbd_send(0x2e); 
#define tecla_6_break  Kbd_send(0xf0); Kbd_send(0x36); 
#define tecla_7_break  Kbd_send(0xf0); Kbd_send(0x3d); 
#define tecla_8_break  Kbd_send(0xf0); Kbd_send(0x3e);
#define tecla_9_break  Kbd_send(0xf0); Kbd_send(0x46); 
#define tecla_a_break  Kbd_send(0xf0); Kbd_send(0x1c); 
#define tecla_b_break  Kbd_send(0xf0); Kbd_send(0x32); 
#define tecla_c_break  Kbd_send(0xf0); Kbd_send(0x21); 
#define tecla_d_break  Kbd_send(0xf0); Kbd_send(0x23); 
#define tecla_e_break  Kbd_send(0xf0); Kbd_send(0x24); 
#define tecla_f_break  Kbd_send(0xf0); Kbd_send(0x2b); 


//codes a ser enviado pelas teclas ao apertar o teclado
#define tecla_0_make  Kbd_send(0x45);//code para letra  '0'
#define tecla_1_make  Kbd_send(0x16);
#define tecla_2_make  Kbd_send(0x1e);
#define tecla_3_make  Kbd_send(0x26);
#define tecla_4_make  Kbd_send(0x25);
#define tecla_5_make  Kbd_send(0x2e); 
#define tecla_6_make  Kbd_send(0x36); 
#define tecla_7_make  Kbd_send(0x3d); 
#define tecla_8_make  Kbd_send(0x3e);
#define tecla_9_make  Kbd_send(0x46); 
#define tecla_a_make  Kbd_send(0x1c); 
#define tecla_b_make  Kbd_send(0x32); 
#define tecla_c_make  Kbd_send(0x21); 
#define tecla_d_make  Kbd_send(0x23); 
#define tecla_e_make  Kbd_send(0x24); 
#define tecla_f_make  Kbd_send(0x2b); 


/*******************************************************************************
 'make code'(apertar) 1byte - scan code (veja tabela)
 'break code'(soltar) 2 byte - 0xf0 + scan code
  Ao usar make code= 'caps locks' dar delay de 100 us antes de enviar break code.
*******************************************************************************/

/*******************************************************************************
             ROTINAS DE CLOCK
*******************************************************************************/
void clock(){ delay_us(8);tris_clk=0;clk=0; delay_us(50);tris_clk=1;delay_us(8);}//20 a 100 us delay do meio
void iclock(){ delay_us(15);tris_clk=1; delay_us(36);tris_clk=0;clk=0;delay_us(15);}//20 a 100 us delay do meio
/*******************************************************************************
             ROTINA PARA ENVIAR UM CDIGO DE TECLA
*******************************************************************************/
int8 envia_code(unsigned int8 code){
  unsigned int8 paridade,b;

  tris_clk=1;//entrada em clk
  if(!clk)return (0xff);//sai da rotina se PC (host) tenha inibido a transmisso

  //envia start bit
  tris_dado=0;
  dado=0;
  clock();
  if(!clk)return (0xff);//sai da rotina se PC (host) tenha inibido a transmisso
  paridade=0;//zera registrador
  //envia 8 bits do 'code'
 for(b=0;b<8;b++){
  if((code&1)==0){ dado=0;} else {dado=1;}
  paridade+=code;//soma todos os bits
  clock();
  if(!clk)return (0xff);//sai da rotina se PC (host) tenha inibido a transmisso
  code=code>>1;
 }
 paridade&=1;//limita a 1 bit e obtem a paridade
 //envia a paridade
 if(paridade==0){ dado=1;} else {dado=0;}
  clock();
  if(!clk)return (0xff);//sai da rotina se PC (host) tenha inibido a transmisso

 //envia o bit de parada
 dado=1;
 clock();
 if(!clk)return (0);//sinaliza sucesso no envio
}
/*******************************************************************************
         ENVIA CODE DO TECLADO PARA O PC (HOST)
******************************************************************************/
void Kbd_send(unsigned int8 dado){
 unsigned int8 buf,cnt;
 #define NUMERO_TENTATIVAS  3  //tenta enviar por 3 vezes
      disable_interrupts(int_ext);
      buf=dado;//code a enviar para pc
      cnt=0;
      while((envia_code(buf)==0xff)&& cnt<NUMERO_TENTATIVAS){
      delay_us(120);//aguarda fim da inibio pelo PC e reenvia
      cnt++;//incrementa contador de tentativas
      }
      enable_interrupts(int_ext);
}
/*******************************************************************************
                   OBTEM CDIGO PARA TECLAS 'MAKE' (APERTANDO TECLA)
*******************************************************************************/
void get_code_make(unsigned int8 f){
   switch (f){

    case 0: {tecla_0_make  break;} //'0'
    case 1: {tecla_1_make  break;} //'1'
    case 2: {tecla_2_make  break;} //'2'
    case 3: {tecla_3_make  break;} //'3'
    case 4: {tecla_4_make  break;} //'4'
    case 5: {tecla_5_make  break;} //'5'
    case 6: {tecla_6_make  break;} //'6'
    case 7: {tecla_7_make  break;} //'7'
    case 8: {tecla_8_make  break;} //'8'
    case 9: {tecla_9_make  break;} //'6'
    case 10: {tecla_a_make  break;} //'a'
    case 11: {tecla_b_make  break;} //'b'
    case 12: {tecla_c_make  break;} //'c'
    case 13: {tecla_d_make  break;} //'d'
    case 14: {tecla_e_make  break;} //'e'
    case 15: {tecla_f_make  break;} //'f'
   }
}
/*******************************************************************************
                   OBTEM CDIGO PARA TECLAS 'BREAK' (SOLTANDO TECLA)
*******************************************************************************/
void get_code_break(unsigned int8 g){
   switch (g){

    case 0: {tecla_0_break  break;} //'0'
    case 1: {tecla_1_break  break;} //'1'
    case 2: {tecla_2_break  break;} //'2'
    case 3: {tecla_3_break  break;} //'3'
    case 4: {tecla_4_break  break;} //'4'
    case 5: {tecla_5_break  break;} //'5'
    case 6: {tecla_6_break  break;} //'6'
    case 7: {tecla_7_break  break;} //'7'
    case 8: {tecla_8_break  break;} //'8'
    case 9: {tecla_9_break  break;} //'9'
    case 10: {tecla_a_break  break;} //'a'
    case 11: {tecla_b_break  break;} //'b'
    case 12: {tecla_c_break  break;} //'c'
    case 13: {tecla_d_break  break;} //'d'
    case 14: {tecla_e_break  break;} //'e'
    case 15: {tecla_f_break  break;} //'f'
   }
}
/*******************************************************************************
         SINALIZA UMA TECLA APERTADA PELO USURIO NO TECLADO 4X4
*******************************************************************************/
#INT_RB
void isr_teclas(){
  f_tec=1;//seta flag de tecla apertada
  portb&=portb;// apenas permite o apagamento do flag de interrupo de mudana
}
/*******************************************************************************
         RECEBE CODE - PC PARA O TECLADO
*******************************************************************************/
#INT_EXT
void recebe_comando(){
       unsigned int8  paridade, par,a;

       disable_interrupts(int_ext);//inibe interrupo externa (no clk pelo PC)

       if(!clk){
        while(!clk); //aguarda at ficar '1'
        if(!dado){
           iclock();//clock inicial para indicar que aceita recepo
           buf=0;
           paridade=0;
           for(a=0;a<8;a++){
             tris_clk=1;//rising no clock
             if(dado){buf|=0x80; paridade++;}
             buf=buf>>1;
             iclock();
           }
        //bit de paridade
           par=0;
           tris_clk=1;//rising no clock
           par=dado;
           iclock();
        // bit de stop
           iclock();
        //bit de acknowledge(confirma para o PC que recebeu o dado)
           dado=0;
           delay_us(30);
           tris_clk=1;
           delay_us(30);
           dado=1;
       //  if((paridade&1)==par)sig=1; else sig=2;//confere paridade (opcional)
           sig=1;//sem teste de paridade
        }// if(!dado){
       }//if(!clk){
      else{  sig=0;}
      enable_interrupts(int_ext);//habilita interrupo externa (no clk pelo PC)
}
/*******************************************************************************
             ROTINA MAIN
*******************************************************************************/
void main() {

   setup_counters(RTCC_INTERNAL,RTCC_DIV_1);//n/usado
   setup_timer_1(T1_DISABLED);//n/usado
   setup_timer_2(T2_DISABLED,0,1);//n/usado
   setup_comparator(NC_NC_NC_NC);//n/usado
   setup_vref(FALSE);//n/usado
   porta=0b11110000;//situao do porta ps reset
   trisa=0b11100000;//entradas e sadas do porta
   trisb=0b11111111;//idem port b
   tris_clk=1;//linha clk=1
   dado=1;//linha de dados=1;
   sig=0;//reseta flag de tecla apertada
   delay_ms(100);//delay
   port_b_pullups(TRUE);//liga pullups no portb
   enable_interrupts(INT_EXT);//liga interrupo externa
   enable_interrupts(INT_RB);//liga interrupo por mudana nos pinos 10 a 13
   enable_interrupts(GLOBAL);//liga chave geral das interrupes
   old_tecla=0xff;//reseta registrador de ultima tecla pressionada
/*******************************************************************************
             LOOP PRINCIPAL DE REPETIO
*******************************************************************************/

   for(;;){

       if(sig){//se '1' sinaliza que h comunicao host--> teclado
          if(buf==0xf2){Kbd_send(0xfa); Kbd_send(0xAB); Kbd_send(0x83);}//envia ID
          else {Kbd_send(0xfa);}//outros comandos apenas acusa recebimento
       }

       if(f_tec){
        disable_interrupts(INT_RB);//inibe interrupo de mudana (teclado)
        disable_interrupts(INT_EXT);//inibe interrupo externa (clk - pelo PC)
        delay_ms(40);//debounce
        tecla=0xff;
        porta|=0x0f;
        porta&=0b11110111;
        delay_us(50);
        if((portb&0xf0)==0b11100000)tecla=15;//MODE
        if((portb&0xf0)==0b11010000)tecla=14;//FLASH
        if((portb&0xf0)==0b10110000)tecla=13;//N/USADO EM TECLADO DE TELEFONE
        if((portb&0xf0)==0b01110000)tecla=12;//FLASH
        porta|=0x0f;
        porta&=0b11111011;
        delay_us(50);
        if((portb&0xf0)==0b11100000)tecla=3;//
        if((portb&0xf0)==0b11010000)tecla=6;//
        if((portb&0xf0)==0b10110000)tecla=9;//
        if((portb&0xf0)==0b01110000)tecla=11;//#
        porta|=0x0f;
        porta&=0b11111101;
        delay_us(50);
        if((portb&0xf0)==0b11100000)tecla=2;//
        if((portb&0xf0)==0b11010000)tecla=5;//
        if((portb&0xf0)==0b10110000)tecla=8;//
        if((portb&0xf0)==0b01110000)tecla=0;//
        porta|=0x0f;
        porta&=0b11111110;
        delay_us(50);
        if((portb&0xf0)==0b11100000)tecla=1;//
        if((portb&0xf0)==0b11010000)tecla=4;//
        if((portb&0xf0)==0b10110000)tecla=7;//
        if((portb&0xf0)==0b01110000)tecla=10;//*

       if(old_tecla!=tecla){
         if(old_tecla==0xff){
         get_code_make(tecla);
         old_tecla=tecla;
         }//make code (tecla apertada)
         else{
         get_code_break(old_tecla);
         old_tecla=0xff; }//break code

       }
       f_tec=0;
       porta&=0xf0;//condio p/ que qualquer tecla que for apertada gere interrupo
       enable_interrupts(INT_RB);//habilita interrupo de mudana (teclado)
       enable_interrupts(INT_EXT);//habilita interrupo externa (clk - pelo PC)
    }

  }// for(;;){
}//main()


